#  include <malloc.h>
#  include <stdio.h>
#  include <sys/types.h>
#  include <sys/time.h>
#  include <pthread.h>
#  include <unistd.h>
#  include <stdlib.h>
#  include <dlfcn.h>
#  include <string.h>
/*
 *  MEM_ARRAY_SIZE defines how many mallocs occur before calls to free are
 *  made.  MEM_SAVE_SIZE defines what number of mallocs are not freed in
 *  the current free cycle.  Increasing the ARRAY_SIZE increases the amount
 *  of memory that is malloced and freed.  Decreasing the SAVE_SIZE divisor
 *  also increases the amount of memory needed to run this test.
 *  This test assumes that 1/15th of the mallocs are not freed on each cycle.
 */
#define MEM_ARRAY_SIZE  10000

void timer(void)
{
  static long long start;
  struct timeval tv;

  if(!start) {
    gettimeofday(&tv, 0);
    start = (tv.tv_sec * 1000) + (tv.tv_usec / 1000);
  } else {
    int mins;
    int secs;
    int msecs;

    long long stop;

    gettimeofday(&tv, 0);
    stop = (tv.tv_sec * 1000) + (tv.tv_usec / 1000);

    stop = stop - start;
    msecs = stop % 1000;
    stop = (stop - msecs) / 1000;
    secs = stop % 60;
    stop = (stop - secs) / 60;
    mins = stop;  

    fprintf(stdout, "Elapsed time: %dm%d.%3.3ds\n", mins, secs, msecs); 
  }
}

void* speed_test(void* param)
{ 
  char** memory;
  int i; 
  size_t j;
  int k;
  int k_max;
  int repeat = 100;
  char** mp;
  char** mpe;
  unsigned long id;
  int max_j = 0;

  memory = (char**)malloc(MEM_ARRAY_SIZE * sizeof(char*) );
  mp  = memory;
  mpe = memory + MEM_ARRAY_SIZE;

  while(repeat--)  
  { 

    /* Report progess when each 20% of the test is completed */
    if(repeat % 20 == 0)
      fprintf(stdout, "thread %ld completed %d%% max size %d\n", pthread_self(), (100 - repeat), max_j);

    for(i = 0; i < 100000; i = i * 3 / 2 + 1)	// i controls the range of sizes
    { 
      for(j = i; j ; j/= 2)			// j counts back down from current i
      {
        k_max = 1;           /* k_max controls the size allocation bias.
                                Large things are allocated once.  Sizes less
                                than 100 are allocated 250 times.
                              */

        if (max_j < j)
          max_j = j;

        if(j < 10000)
          k_max = 10;

        if(j < 1000)
          k_max *= 5;

        if(j < 100)
          k_max *= 10;

        for(k = 0; k < k_max; k ++)
        { 
          if( !(*mp++  = (char*)malloc(j)))
          {
            fprintf(stderr, "? malloc failed ?\n");
            exit(1);
          }

          if(mp >= mpe)   				/* if we've reached the end of the malloc buffer */
          {
            while(mp > memory) {
              free(*--mp);
             }
          }

        }
      }
    }
  }

  /* free the residual allocations */
  while(mp > memory)
    free(*--mp);

  free(memory);
  return (void*)0;
}


#define MAX_THREADS 128
int main(int argc, char* argv[])
{ char* p;
  int thread_count = 1;
  pthread_t threads[MAX_THREADS];
  pthread_t pthread_dummy;
  pthread_attr_t attr;
  int error;
  int i;

  
  if(argc > 1)
    thread_count = atoi(argv[1]);
  if (0 >= thread_count)
   thread_count = 1;
  if (thread_count > MAX_THREADS)
   thread_count = MAX_THREADS;

  
  pthread_attr_init(&attr);
  pthread_attr_setscope(&attr, PTHREAD_SCOPE_SYSTEM);


  timer();

  for (i = 0; i < thread_count; i++) {
    if(0 != (error = pthread_create(&threads[i], &attr, speed_test, (void*)0)))
    { fprintf(stdout, "Cannot start thread, error code: %d\n", error);
      exit(1);
    }
  }

  for(i = 0; i < thread_count; ++i)
    pthread_join(threads[i], (void**)0);

  timer();

  return 0;
}
 
